#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Tuple, List, Dict
import logging
import time
import os
from collections import defaultdict
from celery import Celery, Task
from celery.worker.request import Request
from celery.signals import after_setup_task_logger, after_setup_logger, worker_process_init, \
    celeryd_after_setup, worker_process_shutdown
from celery.utils import worker_direct
from celery.exceptions import SoftTimeLimitExceeded
import billiard
from billiard.exceptions import WorkerLostError
from mongoengine import Document, StringField, ListField, DateTimeField, IntField, ValidationError
from mongoengine.connection import ConnectionFailure
from datetime import datetime
import functools

from nomad import config, utils, infrastructure
import nomad.patch  # pylint: disable=unused-import


if config.logstash.enabled:
    from nomad.utils import structlogging

    def initialize_logstash(logger=None, loglevel=logging.DEBUG, **kwargs):
        structlogging.add_logstash_handler(logger)
        return logger

    after_setup_task_logger.connect(initialize_logstash)
    after_setup_logger.connect(initialize_logstash)


@worker_process_init.connect
def setup(**kwargs):
    # each subprocess is supposed disconnect connect again: https://jira.mongodb.org/browse/PYTHON-2090
    try:
        from mongoengine import disconnect
        disconnect()
    except Exception:
        pass

    infrastructure.setup()
    utils.get_logger(__name__).info(
        'celery configured with acks_late=%s' % str(config.celery.acks_late))


worker_hostname = None


@celeryd_after_setup.connect
def capture_worker_name(sender, instance, **kwargs):
    global worker_hostname
    worker_hostname = sender


@worker_process_shutdown.connect
def on_worker_process_shutdown(*args, **kwargs):
    # We need to make sure not to leave open sessions: https://jira.mongodb.org/browse/PYTHON-2090
    from mongoengine.connection import disconnect
    disconnect()


app = Celery('nomad.processing', broker=config.rabbitmq_url())
app.conf.update(worker_hijack_root_logger=False)
app.conf.update(worker_max_memory_per_child=config.celery.max_memory)
if config.celery.routing == config.CELERY_WORKER_ROUTING:
    app.conf.update(worker_direct=True)

app.conf.task_queue_max_priority = 10


class ProcessStatus:
    '''
    Class holding constants related to the possible process statuses.

    Attributes:
        READY: The process is ready to start
        PENDING: The process has been called, but still waiting for a celery worker to start running.
        RUNNING: Currently running the main process function.
        WAITING_FOR_RESULT: Waiting for the result from some other process.
        SUCCESS: The last process completed successfully.
        FAILURE: The last process completed with a fatal failure.
        DELETED: Used to signal that the process results in the deletion of the object.

        STATUSES_PROCESSING: List of statuses where the process is still incomplete (no other
            process can be started).
        STATUSES_NOT_PROCESSING: The opposite of the above - statuses from which a new
            process can be started.
    '''
    READY = 'READY'
    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    WAITING_FOR_RESULT = 'WAITING_FOR_RESULT'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'
    DELETED = 'DELETED'

    STATUSES_PROCESSING = (PENDING, RUNNING, WAITING_FOR_RESULT)
    STATUSES_NOT_PROCESSING = (READY, SUCCESS, FAILURE)
    STATUSES_COMPLETED = (SUCCESS, FAILURE)
    STATUSES_VALID_IN_DB = tuple(list(STATUSES_NOT_PROCESSING) + list(STATUSES_PROCESSING))


class InvalidId(Exception): pass


class ProcNotRegistered(Exception): pass


class ProcessAlreadyRunning(Exception): pass


class ProcessSyncFailure(Exception): pass


class ProcObjectDoesNotExist(Exception): pass


class ProcessFailure(Exception):
    '''
    Special exception class which allows the user to control how :func:`Proc.fail` should
    be called when the exception is caught from the process function.
    '''
    def __init__(self, *errors, log_level=logging.ERROR, **kwargs):
        self._errors = errors
        self._log_level = log_level
        self._kwargs = kwargs


class Proc(Document):
    '''
    Base class for objects that are subject to processing and need persistent processing
    state. The processing state is persisted in mongo db. Possible processing statuses are
    defined by :class:`ProcessStatus`.

    To initiate a process, an object subclassing Proc must first be created using
    :func:`create`. Processes are then initiated by calling a *process function* on this
    object, which is a member function marked with the decorator @process. Calling a process
    function sets the process_state to PENDING and a celery task is created, which will be
    picked up by a worker, which sets the state to RUNNING and actually executes the
    process function.

    From process_status RUNNING, the process can transition to either SUCCESS, FAILURE or
    WAITING_FOR_RESULT, or result in the deletion of the process object itself (for example
    the process for deleting an upload).

    WAITING_FOR_RESULT means the process needs to wait for the result from spawned child
    processes. When a child process finishes, it will call `try_to_join` on the parent, and
    if it is the parent's last running child process, we will switch to the parent Proc,
    put its status to RUNNING and execute the `join` method on the parent.

    To send the process to status WAITING_FOR_RESULT, the process function must return
    normally, without errors and exceptions, and return the value `ProcessStatus.WAITING_FOR_RESULT`.

    If the process deletes the object itself, the process function should instead return
    `ProcessStatus.DELETED`. If the process function returns normally, without exception
    and without a return value, the process status will be set to SUCCESS.


    Attributes:
        errors: a list of errors that happened during processing.
            NOTE: This value is managed by the framework, do not tamper with this value.
            To fail a process, an exception should be raised.
        warnings: a list of warnings that happened during processing. Warnings do not
            fail a processing run
        last_status_message: A short, human readable message from the current process, with
            information about what the current process is doing, or information about the
            completion (successful or not) of the last process, if no process is currently
            running.
        complete_time: the time that processing completed (successfully or not).
            NOTE: This value is managed by the framework, do not tamper with this value.
        current_process: the currently or last run asyncronous process
            NOTE: This value is managed by the framework, do not tamper with this value.
        process_status: one of the values defined by :class:`ProcessStatus`.
            NOTE: This value is managed by the framework, do not tamper with this value.
        queue: A list defining queued up calls, waiting to be run. Each item is a triple
            of [func_name, args, kwargs].
            NOTE: This value is managed by the framework, do not tamper with this value.
        sync_counter: An integeger, incremented every time a "sync" operation is executed,
            to ensure state consistency and atomicity. There are three types of sync operations:
            when scheduling a process, starting a process, and completing a process.
            NOTE: This value is managed by the framework, do not tamper with this value.
    '''

    meta: Any = {
        'abstract': True,
    }

    complete_time = DateTimeField()

    errors = ListField(StringField())
    warnings = ListField(StringField())
    last_status_message = StringField(default=None)

    current_process = StringField(default=None)
    process_status = StringField(default=None)

    worker_hostname = StringField(default=None)
    celery_task_id = StringField(default=None)

    queue = ListField()
    sync_counter = IntField(default=0)

    @property
    def process_running(self) -> bool:
        '''
        Returns True of an asynchrounous process is currently running (or waiting to run).
        NOTE, the return value will reflect the state when the object was last updated from
        mongo, not neccesarily the current state.
        '''
        return self.process_status in ProcessStatus.STATUSES_PROCESSING

    @classmethod
    def process_running_mongoengine_query(cls):
        ''' Returns a mongoengine query dict (to be used in objects) to find running processes. '''
        return dict(process_status__in=ProcessStatus.STATUSES_PROCESSING)

    @property
    def queue_blocked(self) -> bool:
        '''
        If the queue is blocked (i.e. no new @process can be invoked on this object).
        NOTE, the return value will reflect the state when the object was last updated from
        mongo, not neccesarily the current state.
        '''
        if self.process_status in ProcessStatus.STATUSES_PROCESSING:
            blocking_processes = all_blocking_processes[self.__class__.__name__]
            if self.current_process in blocking_processes:
                return True
            for item in self.queue:
                if item[0] in blocking_processes:
                    return True
        return False

    def get_logger(self):
        process = billiard.current_process()  # pylint: disable=no-member
        worker_id = getattr(process, '_nomad_id', None)
        if worker_id is None:
            worker_id = utils.create_uuid()
            setattr(process, '_nomad_id', worker_id)

        return utils.get_logger(
            'nomad.processing', proc=self.__class__.__name__,
            process=self.current_process, process_status=self.process_status,
            process_worker_id=worker_id)

    @classmethod
    def create(cls, **kwargs):
        ''' Factory method that must be used instead of regular constructor. '''
        assert 'process_status' not in kwargs, \
            ''' do not set the status manually, its managed '''

        self = cls(**kwargs)
        self.process_status = ProcessStatus.READY
        self.save()

        return self

    def reset(
            self, force: bool = False, worker_hostname: str = None,
            process_status: str = ProcessStatus.READY,
            errors: List[str] = [], clear_queue: bool = True):
        '''
        Resets the process status. If force is not set, there must be no currently running process.
        NOTE, use this with care! This should normally only be used manually, to fix processes
        that are "stuck" in status processing, for example if the worker has died etc.
        '''
        assert not self.process_running or force

        self.current_process = None
        self.process_status = process_status
        self.errors = errors
        self.warnings = []
        self.worker_hostname = worker_hostname
        if clear_queue:
            self.queue = []

    @classmethod
    def reset_pymongo_update(
            cls, worker_hostname: str = None,
            process_status=ProcessStatus.READY,
            errors: List[str] = [], clear_queue: bool = True):
        '''
        Returns a pymongo update dict part to reset a Proc.
        NOTE, use this with care! This should normally only be used manually, to fix processes
        that are "stuck" in status processing, for example if the worker has died etc.
        '''
        rv = dict(
            current_process=None, process_status=process_status,
            errors=errors, warnings=[], worker_hostname=worker_hostname)
        if clear_queue:
            rv['queue'] = []
        return rv

    @classmethod
    def get_by_id(cls, id: str, id_field: str):
        try:
            obj = cls.objects(**{id_field: id}).first()
        except ValidationError as e:
            raise InvalidId('%s is not a valid id' % id)
        except ConnectionFailure as e:
            raise e

        if obj is None:
            raise KeyError('%s with id %s does not exist' % (cls.__name__, id))

        return obj

    @classmethod
    def get(cls, obj_id):
        return cls.get_by_id(str(obj_id), 'id')

    @staticmethod
    def log(logger, log_level, msg, **kwargs):
        # TODO there seems to be a bug in structlog, cannot use logger.log
        if log_level == logging.ERROR:
            logger.error(msg, **kwargs)
        elif log_level == logging.WARNING:
            logger.warning(msg, **kwargs)
        elif log_level == logging.INFO:
            logger.info(msg, **kwargs)
        elif log_level == logging.DEBUG:
            logger.debug(msg, **kwargs)
        else:
            logger.critical(msg, **kwargs)

    def fail(self, *errors, log_level=logging.ERROR, complete=True, **kwargs):
        '''
        Used to handle when a process fails. Takes strings or exceptions as args.
        The method logs the error(s), updates `self.errors`, `self.last_status_message`,
        `self.process_status`, and calls :func:`on_fail`, and if `complete` == True (default)
        it also saves the object to mongodb. The positional args define the errors. An
        error should normally be an instance of Exception, if not it will be converted to
        a string.

        NOTE, processes should NOT call this method directly, or tamper with self.errors etc.
        Rather, if something goes wrong in a process, it should raise an exception!
        '''
        assert self.process_running, 'Cannot fail a completed process.'

        failed_with_exception = False

        logger = self.get_logger(**kwargs)
        self.errors = []
        for error in errors:
            if isinstance(error, Exception):
                failed_with_exception = True
                self.errors.append('%s: %s' % (error.__class__.__name__, str(error)))
                Proc.log(
                    logger, log_level, 'process failed with exception',
                    exc_info=error, error=str(error))
            else:
                self.errors.append(str(error))

        self.complete_time = datetime.utcnow()

        if not failed_with_exception:
            errors_str = "; ".join([str(error) for error in errors])
            Proc.log(logger, log_level, 'process failed', errors=errors_str)

        try:
            self.on_fail()
        except Exception as e:
            # Oh my, nothing is going our way today
            Proc.log(logger, logging.ERROR, 'on_fail failed', exc_info=e, error=str(e))

        logger.info('process failed')
        if len(self.errors) > 0:
            self.last_status_message = f'Process {self.current_process} failed: {self.errors[-1]}'

        self.process_status = ProcessStatus.FAILURE
        if complete:
            self._sync_complete_process()

    def warning(self, *warnings, log_level=logging.WARNING, **kwargs):
        ''' Allows to save warnings. Takes strings or exceptions as args. '''
        assert self.process_running

        logger = self.get_logger(**kwargs)

        for warning in warnings:
            warning = str(warning)
            self.warnings.append(warning)
            Proc.log(logger, log_level, 'task with warning', warning=warning)

    def set_last_status_message(self, last_status_message: str):
        ''' Sets the `last_status_message` and saves. '''
        assert self.process_running
        self.last_status_message = last_status_message
        self.save()

    def on_success(self):
        ''' To be called whenever a process is about to transition to status SUCCESS. '''
        pass

    def on_fail(self):
        ''' To be called whenever a process is about to transition to status FAILURE. '''
        pass

    def on_waiting_for_result(self):
        ''' To be called whenever a process is about to transition to status WAITING_FOR_RESULT. '''
        pass

    def block_until_complete(self, interval=0.01):
        '''
        Reloads the process constantly until it sees a completed process (FAILURE or SUCCESS).
        Should be used with care as it can block indefinitely. Just intended for testing
        purposes.
        '''
        self.reload()
        while self.process_running:
            time.sleep(interval)
            self.reload()

    def block_until_complete_or_waiting_for_result(self, interval=0.01):
        '''
        Reloads the process constantly until the process is either complete or in status WAITING_FOR_RESULT.
        Should be used with care as it can block indefinitely. Just intended for testing
        purposes.
        '''
        while self.process_status in (ProcessStatus.PENDING, ProcessStatus.RUNNING):
            time.sleep(interval)
            self.reload()

    def _send_to_worker(self, func_name, *args, **kwargs):
        ''' Invokes a celery task, which will prompt a worker to pick up this Proc object. '''
        self_id = self.id.__str__()
        cls_name = self.__class__.__name__

        queue = None
        if config.celery.routing == config.CELERY_WORKER_ROUTING and self.worker_hostname is not None:
            queue = worker_direct(self.worker_hostname).name

        priority = config.celery.priorities.get('%s.%s' % (cls_name, func_name), 1)

        logger = utils.get_logger(__name__, cls=cls_name, id=self_id, func=func_name)
        logger.debug('calling process function', queue=queue, priority=priority)

        return proc_task.apply_async(
            args=[cls_name, self_id, func_name, args, kwargs],
            queue=queue, priority=priority)

    def __str__(self):
        return 'proc celery_task_id=%s worker_hostname=%s' % (self.celery_task_id, self.worker_hostname)

    def parent(self) -> 'Proc':
        '''
        When running a process marked with `is_child`, and the process completes (i.e. succeeds or
        fails), this method will be invoked by the framework to determine the object's parent Proc.
        '''
        raise NotImplementedError('`parent` not implemented')

    def child_cls(self) -> 'Proc':
        '''
        When running a process which spawns child processes and transitions to WAITING_FOR_RESULT,
        this method will be invoked to determine the "child" class when it is time to try to
        join.
        '''
        raise NotImplementedError('`child_cls` not implemented')

    def _try_to_join(self) -> bool:
        '''
        Called on the parent Proc object to join (resume) the current process.
        For the join to succeed, the following must be fullfilled:
            *   This Proc must have status `WAITING_FOR_RESULT`
            *   No instance of the child class (defined by calling :func:`child_cls`)
                linked to this Proc (if any) must be processing.
        If the join succeeds, the `process_status` will be set to `RUNNING` and True
        will be returned. Otherwise, the method just returns False. The method is written so
        that the join should succeed once and only once.
        '''
        self.reload()
        if self.process_status != ProcessStatus.WAITING_FOR_RESULT:
            self.get_logger().debug('trying to join: not waiting for result')
            return False
        children_processing = self.child_cls().objects(**{
            self.id_field: self.id,
            'process_status__in': ProcessStatus.STATUSES_PROCESSING}).count()
        self.get_logger().debug('trying to join', children_processing=children_processing)

        if not children_processing:
            # We may easily get here multiple times if multiple children finish at the same time
            # To join, we need to read and update the mongo record as a single atomic operation
            old_record = self._get_collection().find_one_and_update(
                {'_id': self.id, 'process_status': ProcessStatus.WAITING_FOR_RESULT},
                {'$set': {'process_status': ProcessStatus.RUNNING}})
            if old_record and old_record['process_status'] == ProcessStatus.WAITING_FOR_RESULT:
                # We managed to update the process_status from WAITING_FOR_RESULT to RUNNING
                # I.e. we've joined!
                self.reload()
                return True
        return False

    def join(self):
        '''
        Override, if applicable, to define what to do when joined, i.e. when the process is
        resumed after all child processes are done. Should return either None (if successful)
        or `ProcessStatus.WAITING_FOR_RESULT` if it again wants to wait for child processes.
        '''
        raise NotImplementedError('`join` not implemented')

    def _sync_schedule_process(self, func_name: str, *args, **kwargs) -> bool:
        '''
        Used to schedule a call to the @process function named `func_name` with the provided
        `args` and `kwargs`. The `args` and `kwargs` need to be json serializable. If this
        object was previously not processing, we set the state to PENDING. We return True
        if the task should be sent to a worker, False otherwise (if a process is already
        running, we should not send the task to the worker right away, instead it should
        be sent to a worker when the preceding process finishes).

        The call will fail and raise a :class:`ProcessAlreadyRunning` if a *blocking process*
        is running or has been added to the queue (blocking processes prevent further requests
        to be queued until they have been completed).

        This is the first of three *sync operations*. These should be atomic and occur in
        sequence, each call fully seeing the relevant changes of the previous operation.
        Because of propagation delays and possible race conditions, we use `sync_counter` to
        detect update collisions. Such collisions should be very unusual, but if they occur
        we reload self and try again, up to 3 times (which *should* be enough to guarantee
        success with virtually absolute certainty). Because we may need to reload, an object
        calling a sync operation should have no unsaved changes.
        '''
        try_counter = 0
        while True:
            if self.queue_blocked:
                raise ProcessAlreadyRunning('A blocking process is running or waiting to run')
            prev_process_running = self.process_running
            mongo_update = {
                '$set': {'sync_counter': self.sync_counter + 1}}
            if prev_process_running:
                # Something else is running, add to queue
                if self.queue:
                    mongo_update['$push'] = {'queue': [func_name, args, kwargs]}
                else:
                    mongo_update['$set'].update(queue=[[func_name, args, kwargs]])
            else:
                # Nothing is running
                mongo_update['$set'].update(
                    process_status=ProcessStatus.PENDING,
                    current_process=func_name,
                    last_status_message='Pending: ' + func_name)
            # Try to update self atomically. Will fail if someone else has managed to write
            # a sync op in between.
            old_record = self._get_collection().find_one_and_update(
                {
                    '$and': [
                        {'_id': self.id},
                        {
                            '$or': [
                                {'sync_counter': self.sync_counter},
                                {'sync_counter': {'$exists': False}}
                            ]
                        }
                    ]
                }, mongo_update)
            try_counter += 1
            if old_record and old_record.get('sync_counter') == self.sync_counter:
                # We have successfully scheduled the process!
                self.reload()
                return not prev_process_running
            # Someone else must have written a sync op (ticked up the sync_counter) in between
            if try_counter >= 3:
                # Three failed attempts - should be virtually impossible!
                raise ProcessSyncFailure('Failed to schedule process too many times - should not happen')
            # Otherwise, sleep, reload, and try again
            time.sleep(0.1)
            self.reload()

    def _sync_start_local_process(self, func_name: str):
        '''
        Used to start a *local* process. If successful, the status transitions to RUNNING
        atomically. The call will fail and raise a :class:`ProcessAlreadyRunning` if any
        other process is currently running.

        This is one of three *sync operations*. See :func:`_sync_schedule_process`
        for more info.
        '''
        try_counter = 0
        while True:
            if self.process_running:
                raise ProcessAlreadyRunning('Another process is running or waiting to run')
            mongo_update = {
                '$set': dict(
                    sync_counter=self.sync_counter + 1,
                    process_status=ProcessStatus.RUNNING,
                    current_process=func_name,
                    last_status_message='Started: ' + func_name,
                    worker_hostname=None,
                    celery_task_id=None,
                    errors=[],
                    warnings=[])}
            # Try to update self atomically. Will fail if someone else has managed to write
            # a sync op in between.
            old_record = self._get_collection().find_one_and_update(
                {
                    '$and': [
                        {'_id': self.id},
                        {
                            '$or': [
                                {'sync_counter': self.sync_counter},
                                {'sync_counter': {'$exists': False}}
                            ]
                        }
                    ]
                }, mongo_update)
            try_counter += 1
            if old_record and old_record.get('sync_counter') == self.sync_counter:
                # We have successfully started the process!
                self.reload()
                return
            # Someone else must have written a sync op (ticked up the sync_counter) in between
            if try_counter >= 3:
                # Three failed attempts - should be virtually impossible!
                raise ProcessSyncFailure('Failed to start local process too many times - should not happen')
            # Otherwise, sleep, reload, and try again
            time.sleep(0.1)
            self.reload()

    def _sync_complete_process(self) -> Tuple[str, List[Any], Dict[str, Any]]:
        '''
        Used to complete a process (when done, successful or not). Returns a triple
        containing information about the next process to run (if any), of the
        form (func_name, args, kwargs).

        There are 3 possibilities:
            1)  There is something in the queue, and the current process was successful
                -> We set status to PENDING and return the next process.
            2)  There is something in the queue, and the current process FAILED:
                -> We clear the queue, set status to FAILURE and return None.
            3)  There is nothing in the queue:
                -> We set the status to the provided value and return None

        This is one of three *sync operations*. See :func:`_sync_schedule_process`
        for more info.
        '''
        assert self.process_status in ProcessStatus.STATUSES_COMPLETED
        # As a safety precausion, save all updates made to the object except the status
        # (We want it to have status RUNNING until the atomic read/write finishes)
        process_status = self.process_status
        self.process_status = ProcessStatus.RUNNING
        self.save()
        self.process_status = process_status

        try_counter = 0
        while True:
            next_process = None
            mongo_update = {'$set': {'sync_counter': self.sync_counter + 1}}
            if self.queue:
                if process_status == ProcessStatus.SUCCESS:
                    next_process = self.queue[0]
                    next_func_name = next_process[0]
                    mongo_update['$pop'] = {'queue': -1}  # pops the first element
                    mongo_update['$set'].update(
                        process_status=ProcessStatus.PENDING,
                        last_status_message='Pending: ' + next_func_name,
                        current_process=next_func_name)
                else:
                    # Failed - clear the queue
                    mongo_update['$set'].update(process_status=process_status, queue=[])
            else:
                mongo_update['$set'].update(process_status=process_status)
            # Try to update self atomically. Will fail if someone else has managed to write
            # a sync op in between.
            old_record = self._get_collection().find_one_and_update(
                {'_id': self.id, 'sync_counter': self.sync_counter}, mongo_update)
            try_counter += 1
            if old_record and old_record.get('sync_counter') == self.sync_counter:
                # We have successfully completed the process
                return next_process
            # Someone else must have written a sync op (ticked up the sync_counter) in between
            if try_counter >= 3:
                # Three failed attempts - should be virtually impossible!
                raise ProcessSyncFailure('Failed to complete process too many times - should not happen')
            # Make another attempt
            time.sleep(0.1)
            self.reload()


def all_subclasses(cls):
    ''' Helper method to calculate set of all subclasses of a given class. '''
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


all_proc_cls = {cls.__name__: cls for cls in all_subclasses(Proc)}
''' Name dictionary for all Proc classes. '''

all_blocking_processes: Dict[str, List[str]] = defaultdict(list)
''' { <proc class name>: <list of all blocking processes defined for this class> } '''


class NomadCeleryRequest(Request):
    '''
    A custom celery request class that allows to catch error in the worker main
    thread, which cannot be caught on the worker threads themselves.
    '''

    def _fail(self, event, **kwargs):
        args = self._payload[0]
        # this might be run in the worker main thread, which does not have a mongo
        # connection by default
        if infrastructure.mongo_client is None:
            infrastructure.setup_mongo()

        proc = unwarp_task(self.task, *args)
        proc.fail(event, **kwargs)

    def on_timeout(self, soft, timeout):
        if not soft:
            self._fail('task timeout occurred', timeout=timeout)

        super().on_timeout(soft, timeout)

    def on_failure(self, exc_info, send_failed_event=True, return_ok=False):
        if isinstance(exc_info.exception, WorkerLostError):
            infrastructure.setup()
            utils.get_logger(__name__).error(
                'detected WorkerLostError', exc_info=exc_info.exception)
            self._fail(
                'process failed due to worker lost: %s' % str(exc_info.exception),
                exc_info=exc_info)

        super().on_failure(
            exc_info,
            send_failed_event=send_failed_event,
            return_ok=return_ok
        )


class NomadCeleryTask(Task):
    Request = NomadCeleryRequest


def unwarp_task(task, cls_name, self_id, *args, **kwargs):
    '''
    Retrieves the proc object that the given task is executed on from the database.
    '''
    logger = utils.get_logger(__name__, cls=cls_name, id=self_id)

    # get the process class
    global all_proc_cls
    cls = all_proc_cls.get(cls_name, None)
    if cls is None:
        # refind all Proc classes, since more modules might have been imported by now
        all_proc_cls = {cls.__name__: cls for cls in all_subclasses(Proc)}
        cls = all_proc_cls.get(cls_name, None)

    if cls is None:
        logger.critical('document not a subcass of Proc')
        raise ProcNotRegistered('document %s not a subclass of Proc' % cls_name)

    # get the process instance
    try:
        try:
            self = cls.get(self_id)
        except KeyError as e:
            from nomad.app import flask
            if flask.app.config['TESTING']:
                # This only happens in tests, where it is not always avoidable that
                # tasks from old test-cases bleed over.
                raise ProcObjectDoesNotExist()
            logger.warning('called object is missing, retry')
            raise task.retry(exc=e, countdown=3)
    except KeyError:
        logger.critical('called object is missing, retries exeeded', proc_id=self_id)
        raise ProcObjectDoesNotExist()

    return self


@app.task(
    bind=True, base=NomadCeleryTask, ignore_results=True, max_retries=3,
    acks_late=config.celery.acks_late, soft_time_limit=config.celery.timeout,
    time_limit=config.celery.timeout * 2)
def proc_task(task, cls_name, self_id, func_name, args, kwargs):
    '''
    The celery task that is used to execute async process functions.
    It retries for 3 times with a countdown of 3 in case of propagation problems, since this
    might happen in sharded, distributed mongo setups where the updates might not
    have yet propagated to everyone.
    '''
    # Obtain the Proc object. Raises exception to make celery retry if object has not propagated.
    proc: Proc = unwarp_task(task, cls_name, self_id)
    logger = proc.get_logger()
    logger.debug('Executing celery task')

    if '_meta_label' in kwargs:
        config.meta.label = kwargs['_meta_label']
        del(kwargs['_meta_label'])

    try_to_join = False
    deleting = False

    # get the process function
    func = getattr(proc, func_name, None)
    if func is None:  # "Should not happen"
        logger.error('called function not a function of proc class')
        proc.fail('called function %s is not a function of proc class %s' % (func_name, cls_name))
        return

    # unwrap the process decorator
    unwrapped_func = getattr(func, '__process_unwrapped', None)
    is_child = getattr(func, '__is_child', False)
    if unwrapped_func is None:  # "Should not happen"
        logger.error('called function was not decorated with @process')
        proc.fail('called function %s was not decorated with @process' % func_name)
        return

    # call the process function
    try:
        os.chdir(config.fs.working_directory)
        with utils.timer(logger, 'process executed on worker', log_memory=True):
            # Set state to RUNNING
            proc.process_status = ProcessStatus.RUNNING
            proc.last_status_message = 'Started: ' + func_name
            proc.worker_hostname = worker_hostname
            proc.celery_task_id = task.request.id
            proc.errors = []
            proc.warnings = []
            proc.save()
            # Actually call the process function
            rv = unwrapped_func(proc, *args, **kwargs)
            if proc.errors:
                # Should be impossible unless the process has tampered with self.errors, which
                # it should not do. We will treat it essentially as if it had raised an exception
                proc.fail('completed with errors but no exception, should not happen', complete=False)
            elif rv is None:
                # All looks good
                proc.on_success()
                proc.process_status = ProcessStatus.SUCCESS
                proc.complete_time = datetime.utcnow()
                if proc.warnings:
                    proc.last_status_message = f'Process {func_name} completed with warnings'
                else:
                    proc.last_status_message = f'Process {func_name} completed successfully'
                logger.info('completed process')
            elif rv == ProcessStatus.WAITING_FOR_RESULT:
                # No errors, and the process requests to wait for other processes
                proc.on_waiting_for_result()
                proc.process_status = ProcessStatus.WAITING_FOR_RESULT
                proc.save()
                # The process could be ready to resume immediately, for example if no child
                # processes were spawned or if they all finished before the parent process.
                try_to_join = True
                logger.info('waiting for results')
            elif rv == ProcessStatus.DELETED:
                # The Proc object itself to be deleted from the database
                deleting = True
            else:
                raise ValueError('Invalid return value from process function')
    except SystemExit as e:
        proc.fail(e)
        return
    except SoftTimeLimitExceeded as e:
        logger.error('exceeded the celery task soft time limit')
        proc.fail(e, complete=False)
    except ProcessFailure as e:
        # Exception with details about how to call self.fail
        proc.fail(*e._errors, log_level=e._log_level, complete=False, **e._kwargs)
    except Exception as e:
        proc.fail(e, complete=False)

    # The proc is done running
    if is_child and proc.process_status in ProcessStatus.STATUSES_COMPLETED:
        # Save and "switch" to the parent to try to join.
        try:
            proc._sync_complete_process()  # Child processes are blocking, so their queues should be empty
            proc = proc.parent()
            logger = proc.get_logger()
            try_to_join = True
        except Exception as e:  # "Should not happen"
            proc.fail(e)
            return

    while try_to_join:
        try_to_join = False
        try:
            joined = proc._try_to_join()
            if joined:
                logger.info('joined')
                rv = proc.join()
                if rv is None:
                    # Succeeded and process is done
                    proc.on_success()
                    proc.process_status = ProcessStatus.SUCCESS
                    proc.complete_time = datetime.utcnow()
                    if proc.warnings:
                        proc.last_status_message = f'Process {proc.current_process} completed with warnings'
                    else:
                        proc.last_status_message = f'Process {proc.current_process} completed successfully'
                    logger.info('completed process')
                elif rv == ProcessStatus.WAITING_FOR_RESULT:
                    # Waiting for result again
                    proc.on_waiting_for_result()
                    proc.process_status = ProcessStatus.WAITING_FOR_RESULT
                    proc.save()
                    # Need to try to join again because, as before, it could be ready to resume immediately
                    logger.info('waiting for results')
                    try_to_join = True
        except SystemExit as e:
            proc.fail(e)
            return
        except SoftTimeLimitExceeded as e:
            logger.error('exceeded the celery task soft time limit')
            proc.fail(e, complete=False)
        except ProcessFailure as e:
            # Exception with details about how to call self.fail
            proc.fail(*e._errors, complete=False, log_level=e._log_level, **e._kwargs)
        except Exception as e:
            proc.fail(e, complete=False)

    if not deleting and proc.process_status in ProcessStatus.STATUSES_COMPLETED:
        # We are about to transition from RUNNING to completed (FAILURE or SUCCESS)
        # But, if something is queued up we should actually go to PENDING instead, and
        # trigger celery again
        try:
            next_process = proc._sync_complete_process()
            if next_process:
                func_name, args, kwargs = next_process
                proc._send_to_worker(func_name, *args, **kwargs)
        except Exception as e:
            proc.fail(e)


def process(is_blocking: bool = False, is_child: bool = False):
    '''
    The decorator for process functions that will be queued up and executed async via celery.
    To transfer state, the instance will be saved to the database and loading on
    the celery task worker. Process methods can call other (process) functions/methods on
    other :class:`Proc` instances. Each :class:`Proc` instance can only run one process
    at a time. The Proc object should not have any unsaved changes when a process is invoked.

    Arguments:
        is_blocking:
            If True, this is a *blocking process*. After a blocking process has been scheduled,
            no other processes can be scheduled, until this process has finished. Attempts to
            invoke another process will in this case result in an exception.
        is_child:
            If this is a child process, which should try to join with the parent process when done.
            Child processes are implicitly blocking.
    '''
    if is_child:
        is_blocking = True

    def process_decorator(func):
        # Determine canonical class name
        cls_name, func_name = func.__qualname__.split('.')
        if is_blocking:
            all_blocking_processes[cls_name].append(func_name)

        @functools.wraps(func)
        def wrapper(self: Proc, *args, **kwargs):
            kwargs['_meta_label'] = config.meta.label
            send_to_worker = self._sync_schedule_process(func_name, *args, **kwargs)
            if send_to_worker:
                self._send_to_worker(func_name, *args, **kwargs)

        setattr(wrapper, '__process_unwrapped', func)
        setattr(wrapper, '__is_child', is_child)
        return wrapper
    return process_decorator


def process_local(func):
    '''
    The decorator for functions that process locally. These work similarly to functions
    marked with the `@process` decorator, but they are executed directly, in the current
    thread, not via celery. Consequently, they can only be started if no other process is
    running. They are also implicitly blocking, i.e. while running, no other process
    (local or celery-based) can be started or scheduled. If successful, a local process can
    return a value to the caller (unlike celery processes). Invoking a local process should
    only throw exceptions if it was not possible to start the process (because some other
    process is running). Any other errors that happen during the running of the process are
    handled in the usual way, by setting self.errors etc. The Proc object should not have
    any unsaved changes when a local process is invoked.
    '''
    # Determine canonical class name
    cls_name, func_name = func.__qualname__.split('.')
    all_blocking_processes[cls_name].append(func_name)

    def wrapper(self: Proc, *args, **kwargs):
        logger = self.get_logger()
        logger.debug('Executing local process')
        self._sync_start_local_process(func_name)

        try:
            os.chdir(config.fs.working_directory)
            with utils.timer(logger, 'process executed locally', log_memory=True):
                # Actually call the process function
                rv = func(self, *args, **kwargs)
                if self.errors:
                    # Should be impossible unless the process has tampered with self.errors, which
                    # it should not do. We will treat it essentially as if it had raised an exception
                    self.fail('completed with errors but no exception, should not happen', complete=False)
                else:
                    # All looks good
                    self.on_success()
                    self.process_status = ProcessStatus.SUCCESS
                    self.complete_time = datetime.utcnow()
                    if self.warnings:
                        self.last_status_message = f'Process {func_name} completed with warnings'
                    else:
                        self.last_status_message = f'Process {func_name} completed successfully'
                    logger.info('completed process')
                    return rv
        except SystemExit as e:
            self.fail(e, complete=False)
        except ProcessFailure as e:
            # Exception with details about how to call self.fail
            self.fail(*e._errors, log_level=e._log_level, complete=False, **e._kwargs)
        except Exception as e:
            self.fail(e, complete=False)
        finally:
            self._sync_complete_process()  # Queue should be empty, so nothing more to do

    return wrapper



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)
from nomad.datamodel.metainfo import simulation
from nomad.datamodel.metainfo import workflow


m_package = Package()


class x_onetep_section_vibrational_frequencies(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_vibrationl_frequencies = Quantity(
        type=np.dtype(np.float64),
        shape=['len(self.nr_iter)'],
        description='''
        Vibration Frequenices (cm-1)
        ''')

    x_onetep_vibrationl_frequencies_store = Quantity(
        type=str,
        shape=[],
        description='''
        Vibration Frequenices (cm-1)
        ''')

    x_onetep_ir_store = Quantity(
        type=str,
        shape=[],
        description='''
        Irreducible representation in the Point Group
        ''')

    x_onetep_ir = Quantity(
        type=str,
        shape=['len(self.nr_iter)'],
        description='''
        Irreducible representation in the Point Group
        ''')

    x_onetep_raman_activity = Quantity(
        type=np.dtype(np.float64),
        shape=['len(self.nr_iter)'],
        description='''
        Raman activity (A**4/amu)
        ''')

    x_onetep_raman_active = Quantity(
        type=str,
        shape=[],
        description='''
        Raman activity (A**4/amu)
        ''')

    x_onetep_n_iterations_phonons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of iterations in phonons
        ''')

    x_onetep_raman_activity_store = Quantity(
        type=str,
        shape=[],
        description='''
        Raman activity (A**4/amu)
        ''')

    x_onetep_ir_intensity = Quantity(
        type=np.dtype(np.float64),
        shape=['len(self.nr_iter)'],
        description='''
        IR intensities (D/A)**2/amu
        ''')

    x_onetep_ir_intensity_store = Quantity(
        type=str,
        shape=[],
        description='''
        IR intensities (D/A)**2/amu
        ''')


class x_onetep_section_tddft(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_tddft_energies = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        lrtddft energy
        ''')

    x_onetep_tddft_penalties = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        lrtddft energy
        ''')

    x_onetep_tddft_steps = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        lrtddft energy
        ''')

    x_onetep_tddft_omega_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        change in omega
        ''')

    x_onetep_tddft_rms_gradient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        rms gradient tddft
        ''')

    x_onetep_tddft_number_conv_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number converged states
        ''')

    x_onetep_tddft_excit_energies = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        excitations
        ''')

    x_onetep_tddft_excit_oscill_str = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        oscillator strenght
        ''')

    x_onetep_tddft_excit_lifetime = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        excit_lifetime
        ''')

    x_onetep_section_tddft_excitations = SubSection(
        sub_section=SectionProxy('x_onetep_section_tddft_excitations'),
        repeats=True)

    x_onetep_section_tddft_iterations = SubSection(
        sub_section=SectionProxy('x_onetep_section_tddft_iterations'),
        repeats=True)


class x_onetep_section_tddft_excitations(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_tddft_excit_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        exciations
        ''')

    x_onetep_tddft_excit_oscill_str = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        oscillator strenght
        ''')

    x_onetep_tddft_excit_lifetime = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='s',
        description='''
        excit_lifetime
        ''')


class x_onetep_section_tddft_iterations(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_tddft_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lrtddft energy
        ''')

    x_onetep_tddft_penalties = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lrtddft energy
        ''')

    x_onetep_tddft_step = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lrtddft energy
        ''')


class x_onetep_section_atom_ionic_velocities(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)


class x_onetep_section_atom_positions_optim(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_cell_angle_alpha_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle alpha
        ''')

    x_onetep_cell_angle_beta_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle beta
        ''')

    x_onetep_cell_angle_gamma_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle gamma
        ''')

    x_onetep_cell_length_a_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        a unit cell edge length
        ''')

    x_onetep_cell_length_b_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        b unit cell edge length
        ''')

    x_onetep_cell_length_c_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        c unit cell edge length
        ''')


class x_onetep_section_atom_positions(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_cell_angle_alpha = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle alpha
        ''')

    x_onetep_cell_angle_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle beta
        ''')

    x_onetep_cell_angle_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle gamma
        ''')

    x_onetep_cell_length_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        a unit cell edge length
        ''')

    x_onetep_cell_length_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        b unit cell edge length
        ''')

    x_onetep_cell_length_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        c unit cell edge length
        ''')


class x_onetep_section_dipole(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_electronic_dipole_moment_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        store dipole
        ''')

    x_onetep_electronic_dipole_moment = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        -
        ''')

    x_onetep_electronic_dipole_moment_magnitude = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_ionic_dipole_moment_magnitude = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_total_dipole_moment_magnitude = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_ionic_dipole_moment_store = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        -
        ''')

    x_onetep_ionic_dipole_moment = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        -
        ''')

    x_onetep_total_dipole_moment_store = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        -
        ''')

    x_onetep_total_dipole_moment = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        -
        ''')


class x_onetep_section_cell_optim(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_cell_vector_optim = Quantity(
        type=str,
        shape=[],
        description='''
        Temporay storage for cell vectors
        ''')


class x_onetep_section_cell(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_units = Quantity(
        type=str,
        shape=[],
        description='''
        units
        ''')

    x_onetep_cell_vector = Quantity(
        type=str,
        shape=[],
        description='''
        Temporay storage for cell vectors
        ''')


class x_onetep_section_collect_scf_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)


class x_onetep_section_eigenvalues_1(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_eigenvalues_1 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storing eigenvalues
        ''')


class x_onetep_section_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storing eigenvalues
        ''')


class x_onetep_section_functional_definition(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_functional_type = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional definition in onetep convention
        ''')

    x_onetep_functional_weight = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional weight in onetep convention
        ''')


class x_onetep_section_functionals(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_functional_name = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional in onetep convention
        ''')


class x_onetep_section_mulliken_population_analysis(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_mulliken_atom_index = Quantity(
        type=str,
        shape=['number_of_atoms'],
        description='''
        Mulliken_atom_index
        ''')

    x_onetep_mulliken_atom = Quantity(
        type=str,
        shape=['number_of_atoms'],
        description='''
        Mulliken_atom kind
        ''')

    x_onetep_total_orbital = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_total_contribution
        ''')

    x_onetep_total_orbital_store = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_total_contribution
        ''')

    x_onetep_spin = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_spin
        ''')

    x_onetep_spin_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Mulliken_spin
        ''')

    x_onetep_mulliken_charge_store = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_charges
        ''')

    x_onetep_mulliken_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_charges
        ''')


class x_onetep_section_nbo_population_analysis(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_nbo_atom_label_store = Quantity(
        type=str,
        shape=['number_of_atoms'],
        description='''
        nbo_atom kind
        ''')

    x_onetep_nbo_atom_labels = Quantity(
        type=str,
        shape=['number_of_atoms'],
        description='''
        nbo_atom kind
        ''')

    x_onetep_total_nbo_population_store = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        nbo_population
        ''')

    x_onetep_total_nbo_population = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        nbo_population
        ''')

    x_onetep_nbo_partial_charge_store = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        nbo_charges
        ''')

    x_onetep_nbo_partial_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        nbo_charges
        ''')

    x_onetep_nbo_total_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        nbo_charges
        ''')


class x_onetep_section_geom_optimisation_method(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_geometry_optim_method = Quantity(
        type=str,
        shape=[],
        description='''
        Determines optimisation method used
        ''')


class x_onetep_section_optics_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_optics_n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        optics_number_of_bands
        ''')

    x_onetep_optics_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        optics_band_convergence_tolerance
        ''')


class x_onetep_section_ngwf_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ngwf_cg_max_step = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ngwf_cg_max_step
        ''')

    x_onetep_ngwf_cg_type = Quantity(
        type=str,
        shape=[],
        description='''
        ngwf_cg_type
        ''')

    x_onetep_ngwf_halo = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ngwf_halo
        ''')

    x_onetep_ngwf_max_grad = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_onetep_ngwf_max_grad
        ''')

    x_onetep_ngwf_threshold_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ngwf_threshold_orig
        ''')


class x_onetep_section_kernel_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_kernel_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_kernel_diis_maxit = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')

    x_onetep_kernel_diis_coeff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_kernel_diis_thershold = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_kernel_diis_size = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_kernel_diis_type_store = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_onetep_kernel_diis_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_onetep_kernel_diis_lshift = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the shift in energy of the conduction bands with the level-shifting
        technique during the inner loop DIIS. Reference:
        ''')

    x_onetep_kernel_diis_linear_iter = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')


class x_onetep_section_phonons(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_phonon_method = Quantity(
        type=str,
        shape=[],
        description='''
        Phonon calculation method
        ''')

    x_onetep_DFPT_solver_method = Quantity(
        type=str,
        shape=[],
        description='''
        Phonon DFPT solver method
        ''')

    x_onetep_phonon_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon calculation tolerance (eV/A**2)
        ''')

    x_onetep_phonon_cycles = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon calculation cycles
        ''')

    x_onetep_band_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon band convergence tolerance window
        ''')


class x_onetep_section_density_mixing_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_density_mixing_scheme = Quantity(
        type=str,
        shape=[],
        description='''
        density_mixing_scheme
        ''')

    x_onetep_density_mixing_length = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        density_mixing_scheme_length
        ''')

    x_onetep_charge_density_mixing_amplitude = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge_density_mixing_amplitude
        ''')

    x_onetep_cut_off_energy_for_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge_density_mixing_cut_off_energy_for_mixing (A)
        ''')


class x_onetep_section_population_analysis_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_population_analysis_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Population_analysis_cutoff_(A)
        ''')


class x_onetep_section_kernel_optimisation(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total_energy(Eh)
        ''')

    x_onetep_total_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total_energy(Eh)
        ''')

    x_onetep_band_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        band_gap
        ''')

    x_onetep_rms_occupancy_error = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        rms_occupancy
        ''')

    x_onetep_commutator = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Commutator between Hamiltonian and Kernel
        ''')


class x_onetep_section_k_band(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)


class x_onetep_section_k_points_1(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_k_points_1 = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing k points coordinates (fractional)
        ''')


class x_onetep_section_k_points(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_k_points = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing k points coordinates (fractional)
        ''')


class x_onetep_section_relativity_treatment(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_relativity_treatment_scf = Quantity(
        type=str,
        shape=[],
        description='''
        Relativity treatment in onetep convention
        ''')


class x_onetep_section_scf_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_scf_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')


class x_onetep_section_SCF_iteration_frame(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_frame_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        onetep_store_t_md_frame
        ''')

    x_onetep_SCF_frame_energy = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_iterations
        ''')

    x_onetep_SCF_frame_energy_gain = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_iterations_gain
        ''')

    x_onetep_frame_time_scf_iteration_wall_end = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_wall_end_time
        ''')

    x_onetep_frame_energy_free = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        energy_free
        ''')

    x_onetep_frame_energy_total_T0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        energy_free_corrected_for_finite_basis
        ''')


class x_onetep_section_scf_k_points(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_scf_k_points = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_onetep_section_scf_eigenvalues = SubSection(
        sub_section=SectionProxy('x_onetep_section_scf_eigenvalues'),
        repeats=True)


class x_onetep_section_spin_number(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_spin_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')


class x_onetep_section_stress_tensor(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_stress_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing stress tensor components
        ''')


class x_onetep_section_time(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)


class x_onetep_section_raman_tensor(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_store_raman_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged Raman susceptibility tensor
        ''')

    x_onetep_ramen_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='ampere / unified_atomic_mass_unit',
        description='''
        Ramen tensor, unit scaled by 0.5
        ''')


class x_onetep_section_tddft_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_lr_tddft_cg_threshold = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_onetep_lr_tddft_cg_threshold
        ''')

    x_onetep_lr_tddft_kernel_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_onetep_lr_tddft_kernel_cutoff
        ''')

    x_onetep_lr_tddft_maxit_cg = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_onetep_lr_tddft_maxit_cg
        ''')

    x_onetep_lr_tddft_maxit_pen = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_onetep_lr_tddft_maxit_pen
        ''')

    x_onetep_lr_tddft_num_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_onetep_lr_tddft_num_states
        ''')

    x_onetep_lr_tddft_penalty_tol = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_onetep_lr_tddft_penalty_tol
        ''')


class x_onetep_section_scf_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_energy_threshold_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy Threshold store
        ''')

    x_onetep_max_iter_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of maximum iterations steps store
        ''')

    x_onetep_smearing_kind = Quantity(
        type=str,
        shape=[],
        description='''
        Smearing kind
        ''')

    x_onetep_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Smearing width
        ''')

    x_onetep_elec_cg_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_onetep_elec_cg_max
        ''')

    x_onetep_elec_force_tol = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_onetep_elec_force_tol
        ''')


class x_onetep_section_electronic_spectroscpy_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_theory_level = Quantity(
        type=str,
        shape=[],
        description='''
        Electronic spectroscopy parameters theory level
        ''')

    x_onetep_spectroscopy_calculation = Quantity(
        type=str,
        shape=[],
        description='''
        Electronic spectroscopy parameters calculation type
        ''')

    x_onetep_spec_max_iter = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of iterations
        ''')

    x_onetep_spec_max_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of steps
        ''')

    x_onetep_spec_max_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of bands
        ''')

    x_onetep_spec_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Tolerance (eV)
        ''')


class x_onetep_section_md(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_md_energies = Quantity(
        type=str,
        shape=[],
        description='''
        md_energy_components
        ''')

    x_onetep_md_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_temp(K)
        ''')

    x_onetep_md_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_pressure
        ''')

    x_onetep_md_cell_vectors = Quantity(
        type=str,
        shape=[],
        description='''
        md_cell_vectors
        ''')

    x_onetep_md_cell_vectors_vel = Quantity(
        type=str,
        shape=[],
        description='''
        md_cell_vectors_velocities
        ''')

    x_onetep_md_stress_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        md_stress_tensor
        ''')

    x_onetep_md_positions = Quantity(
        type=str,
        shape=[],
        description='''
        md_positions
        ''')

    x_onetep_md_forces = Quantity(
        type=str,
        shape=[],
        description='''
        md_forces
        ''')

    x_onetep_md_lab = Quantity(
        type=str,
        shape=[],
        description='''
        md_lables
        ''')

    x_onetep_md_veloc = Quantity(
        type=str,
        shape=[],
        description='''
        md_veloc
        ''')


class x_onetep_section_edft_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_edft_commutator_thres = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        edft_commutator_thres
        ''')

    x_onetep_edft_energy_thres = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        edft_energy_thres
        ''')

    x_onetep_edft_entropy_thres = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        edft_entropy_thres
        ''')

    x_onetep_edft_extra_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        edft_extra_bands
        ''')

    x_onetep_edft_fermi_thres = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        edft_fermi_thres
        ''')

    x_onetep_edft_free_energy_thres = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        edft_free_energy_thres
        ''')

    x_onetep_edft_maxit = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        edft_maxit
        ''')

    x_onetep_edft_rms_gradient_thres = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        edft_rms_gradient_thres
        ''')

    x_onetep_edft_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        edft_smearing_width
        ''')


class x_onetep_section_ts(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ts_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_total
        ''')

    x_onetep_ts_cell_vectors = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_cell_vectors
        ''')

    x_onetep_ts_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_positions
        ''')

    x_onetep_ts_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_forces
        ''')

    x_onetep_ts_lab = Quantity(
        type=str,
        shape=[],
        description='''
        ts_lables
        ''')


class x_onetep_section_ts_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ts_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_ts_store
        ''')

    x_onetep_ts_cell_vectors_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors
        ''')

    x_onetep_ts_positions_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_positions_store
        ''')

    x_onetep_ts_forces_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_forces_store
        ''')


class x_onetep_section_ts_final_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ts_energy_final_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_ts
        ''')

    x_onetep_ts_forces_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_forces_final
        ''')

    x_onetep_ts_positions_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_positions_final
        ''')

    x_onetep_ts_cell_vectors_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors_final
        ''')


class x_onetep_section_ts_final(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ts_energy_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_final
        ''')

    x_onetep_ts_cell_vectors_final = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_cell_vectors
        ''')

    x_onetep_ts_positions_final = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_positions_final
        ''')

    x_onetep_ts_forces_final = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_force_finals
        ''')


class x_onetep_section_ts_product(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ts_energy_product = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_prod
        ''')

    x_onetep_ts_forces_product = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_energy_forces_pro
        ''')

    x_onetep_ts_positions_product = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_energy_positions_pro
        ''')

    x_onetep_ts_cell_vectors_product = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_product_cell_vectors
        ''')


class x_onetep_section_ts_product_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ts_energy_product_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_pro
        ''')

    x_onetep_ts_forces_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_forces_pro_store
        ''')

    x_onetep_ts_positions_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_positions_pro_store
        ''')

    x_onetep_ts_cell_vectors_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors_pro
        ''')


class x_onetep_section_ts_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_ts_method = Quantity(
        type=str,
        shape=[],
        description='''
        ts_method
        ''')

    x_onetep_ts_protocol = Quantity(
        type=str,
        shape=[],
        description='''
        ts_protocol
        ''')

    x_onetep_ts_qst_iter = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ts_qst_iterations
        ''')

    x_onetep_ts_number_cg = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ts_number_of_cg_iterations
        ''')

    x_onetep_ts_force_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_force_tolerance (eV/A)
        ''')

    x_onetep_ts_displacement_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_displacement_tolerance (A)
        ''')


class x_onetep_section_van_der_Waals_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_disp_method_name_store = Quantity(
        type=str,
        shape=[],
        description='''
        Name type
        ''')

    x_onetep_disp_method_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name type
        ''')

    x_onetep_Parameter_d = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method G06
        ''')

    x_onetep_Parameter_LAMBDA = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method OBS
        ''')

    x_onetep_Parameter_n = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method OBS
        ''')

    x_onetep_Parameter_s6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method G06
        ''')

    x_onetep_Parameter_sR = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method TS
        ''')


class x_onetep_section_energy_components(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_pseudo_local_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Pseudopotential
        ''')

    x_onetep_electronic_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Kinetic component store
        ''')

    x_onetep_energy_correction_hartree_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Hartree correction store
        ''')

    x_onetep_energy_XC_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        exchange energy store
        ''')

    x_onetep_ewald_correction_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ewald correction store
        ''')

    x_onetep_dispersion_correction_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dispersion correction store
        ''')

    x_onetep_integrated_density_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        store_integrated density
        ''')

    x_onetep_pseudo_non_local_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Pseudopotential
        ''')


class x_onetep_section_edft(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_section_edft_iterations = SubSection(
        sub_section=SectionProxy('x_onetep_section_edft_iterations'),
        repeats=True)


class x_onetep_section_edft_spin(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_edft_spin_type = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        spin type
        ''')

    x_onetep_edft_n_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        integrated number of electrons
        ''')

    x_onetep_edft_fermi_level = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        fermi level
        ''')

    x_onetep_edft_fermi_level_delta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        fermi level
        ''')

    x_onetep_section_edft_spin_iterations = SubSection(
        sub_section=SectionProxy('x_onetep_section_edft_spin_iterations'),
        repeats=True)


class x_onetep_section_edft_spin_iterations(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_edft_orbital_iteration_spin = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        orbital
        ''')

    x_onetep_edft_eigenvalue = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        eigenvalues
        ''')

    x_onetep_edft_occupancy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        occupancy
        ''')


class x_onetep_section_edft_iterations(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_edft_step = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        step
        ''')

    x_onetep_edft_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        step
        ''')

    x_onetep_edft_0K = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        step
        ''')

    x_onetep_residual_nonorthog = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        step
        ''')

    x_onetep_residual_n_elec = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        step
        ''')

    x_onetep_edft_commutator = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Commutator
        ''')

    x_onetep_edft_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        free energy A=E-TS
        ''')

    x_onetep_edft_rms_gradient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        rms gradient in edft
        ''')

    x_onetep_edft_entropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        rms gradient in edft
        ''')

    x_onetep_edft_iteration = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        iteration number
        ''')

    x_onetep_section_edft_spin = SubSection(
        sub_section=SectionProxy('x_onetep_section_edft_spin'),
        repeats=True)


class x_onetep_section_orbital_information(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_onetep_total_number_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')

    x_onetep_total_number_occ_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')

    x_onetep_occupancy_sum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_homo_lumo_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_mid_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_orbital_number = Quantity(
        type=np.dtype(np.int32),
        shape=['x_onetep_total_number_occ_orbitals'],
        description='''
        -
        ''')

    x_onetep_orbital_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_onetep_total_number_occ_orbitals'],
        description='''
        -
        ''')

    x_onetep_orbital_occupancy_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_orbital_number_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')

    x_onetep_orbital_energy_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_orbital_occupancy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_onetep_total_number_occ_orbitals'],
        description='''
        -
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_atom_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Forces on the atoms as minus gradient of energy_total, including forces' unitary-
        transformation (rigid body) filtering and including constraints, if present. The
        derivatives with respect to displacements of the nuclei in the gradient are
        evaluated according to the coordinate system defined in coordinate_system. In
        addition, these forces are obtained by filtering out the unitary transformations
        (translations of the center of mass and rigid rotations of the whole system, when
        non periodic), atom_forces_raw for the unfiltered counterpart. Furthermore, forces
        due to constraints like fixed atoms, distances, angles, dihedrals, and so on, are
        here included (see atom_forces_raw for the unfiltered counterpart).
        ''')

    x_onetep_atom_ionforces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        forces
        ''')

    x_onetep_atom_local_potentialforces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        forces
        ''')

    x_onetep_atom_nonlocal_potentialforces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        forces
        ''')

    x_onetep_atom_nonself_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        forces
        ''')

    x_onetep_atom_correction_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        forces
        ''')

    x_onetep_basis_set_planewave_cutoff_iteration_0 = Quantity(
        type=str,
        shape=[],
        description='''
        cutoff at iteration 0 of geometry optimisation
        ''')

    x_onetep_final_time = Quantity(
        type=str,
        shape=[],
        description='''
        onetep_calculation_time
        ''')

    x_onetep_final_date = Quantity(
        type=str,
        shape=[],
        description='''
        onetep_calculation_date
        ''')

    x_onetep_compiler = Quantity(
        type=str,
        shape=[],
        description='''
        Compiler name
        ''')

    x_onetep_constants_reference = Quantity(
        type=str,
        shape=[],
        description='''
        Fundamental constant data source
        ''')

    x_onetep_fft_library = Quantity(
        type=str,
        shape=[],
        description='''
        fft library name
        ''')

    x_onetep_avarage_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        onetep_finalisation_time
        ''')

    x_onetep_geom_converged = Quantity(
        type=str,
        shape=[],
        description='''
        onetep_geom_converged
        ''')

    x_onetep_pbc_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        onetep_pbc cutoff for effect of open boundary
        ''')

    x_onetep_is_smearing = Quantity(
        type=str,
        shape=[],
        description='''
        Turns on the smeared ion representation for electrostatics calculation.
        ''')

    x_onetep_total_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_onetep_Initialisation_time
        ''')

    x_onetep_maths_library = Quantity(
        type=str,
        shape=[],
        description='''
        Maths library name
        ''')

    x_onetep_program_compilation_date = Quantity(
        type=str,
        shape=[],
        description='''
        Compilation date (string)
        ''')

    x_onetep_program_compilation_time = Quantity(
        type=str,
        shape=[],
        description='''
        Compilation time (string)
        ''')

    x_onetep_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        Run execution date (string)
        ''')

    x_onetep_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        Run execution start time (string)
        ''')

    x_onetep_crystal_point_group = Quantity(
        type=str,
        shape=[],
        description='''
        Point group of the crystal (Schoenflies notation)
        ''')

    x_onetep_space_group = Quantity(
        type=str,
        shape=[],
        description='''
        Point space of the crystal
        ''')

    x_onetep_ts_path = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_number
        ''')

    x_onetep_ts_path_ts_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_final
        ''')

    x_onetep_ts_path_product = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_pro
        ''')

    x_onetep_store_atom_forces_band = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_onetep_number_of_processors = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of processors
        ''')

    x_onetep_section_vibrational_frequencies = SubSection(
        sub_section=SectionProxy('x_onetep_section_vibrational_frequencies'),
        repeats=True)

    x_onetep_section_tddft = SubSection(
        sub_section=SectionProxy('x_onetep_section_tddft'),
        repeats=True)

    x_onetep_section_dipole = SubSection(
        sub_section=SectionProxy('x_onetep_section_dipole'),
        repeats=True)

    x_onetep_section_collect_scf_eigenvalues = SubSection(
        sub_section=SectionProxy('x_onetep_section_collect_scf_eigenvalues'),
        repeats=True)

    x_onetep_section_mulliken_population_analysis = SubSection(
        sub_section=SectionProxy('x_onetep_section_mulliken_population_analysis'),
        repeats=True)

    x_onetep_section_nbo_population_analysis = SubSection(
        sub_section=SectionProxy('x_onetep_section_nbo_population_analysis'),
        repeats=True)

    x_onetep_section_geom_optimisation_method = SubSection(
        sub_section=SectionProxy('x_onetep_section_geom_optimisation_method'),
        repeats=True)

    x_onetep_section_optics_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_optics_parameters'),
        repeats=True)

    x_onetep_section_ngwf_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_ngwf_parameters'),
        repeats=True)

    x_onetep_section_kernel_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_kernel_parameters'),
        repeats=True)

    x_onetep_section_phonons = SubSection(
        sub_section=SectionProxy('x_onetep_section_phonons'),
        repeats=True)

    x_onetep_section_density_mixing_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_density_mixing_parameters'),
        repeats=True)

    x_onetep_section_population_analysis_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_population_analysis_parameters'),
        repeats=True)

    x_onetep_section_kernel_optimisation = SubSection(
        sub_section=SectionProxy('x_onetep_section_kernel_optimisation'),
        repeats=True)

    x_onetep_section_SCF_iteration_frame = SubSection(
        sub_section=SectionProxy('x_onetep_section_SCF_iteration_frame'),
        repeats=True)

    x_onetep_section_time = SubSection(
        sub_section=SectionProxy('x_onetep_section_time'),
        repeats=True)

    x_onetep_section_raman_tensor = SubSection(
        sub_section=SectionProxy('x_onetep_section_raman_tensor'),
        repeats=True)

    x_onetep_section_tddft_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_tddft_parameters'),
        repeats=True)

    x_onetep_section_scf_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_scf_parameters'),
        repeats=True)

    x_onetep_section_electronic_spectroscpy_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_electronic_spectroscpy_parameters'),
        repeats=True)

    x_onetep_section_md = SubSection(
        sub_section=SectionProxy('x_onetep_section_md'),
        repeats=True)

    x_onetep_section_edft_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_edft_parameters'),
        repeats=True)

    x_onetep_section_ts = SubSection(
        sub_section=SectionProxy('x_onetep_section_ts'),
        repeats=True)

    x_onetep_section_ts_store = SubSection(
        sub_section=SectionProxy('x_onetep_section_ts_store'),
        repeats=True)

    x_onetep_section_ts_final_store = SubSection(
        sub_section=SectionProxy('x_onetep_section_ts_final_store'),
        repeats=True)

    x_onetep_section_ts_final = SubSection(
        sub_section=SectionProxy('x_onetep_section_ts_final'),
        repeats=True)

    x_onetep_section_ts_product = SubSection(
        sub_section=SectionProxy('x_onetep_section_ts_product'),
        repeats=True)

    x_onetep_section_ts_product_store = SubSection(
        sub_section=SectionProxy('x_onetep_section_ts_product_store'),
        repeats=True)

    x_onetep_section_ts_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_ts_parameters'),
        repeats=True)

    x_onetep_section_van_der_Waals_parameters = SubSection(
        sub_section=SectionProxy('x_onetep_section_van_der_Waals_parameters'),
        repeats=True)

    x_onetep_section_energy_components = SubSection(
        sub_section=SectionProxy('x_onetep_section_energy_components'),
        repeats=True)

    x_onetep_section_edft = SubSection(
        sub_section=SectionProxy('x_onetep_section_edft'),
        repeats=True)

    x_onetep_section_orbital_information = SubSection(
        sub_section=SectionProxy('x_onetep_section_orbital_information'),
        repeats=True)


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_net_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Net charge of system
        ''')

    x_onetep_number_of_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of bands
        ''')

    x_onetep_units_atom_position = Quantity(
        type=str,
        shape=[],
        description='''
        units
        ''')

    x_onetep_number_of_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of electrons
        ''')

    x_onetep_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Storing atomic positions in fractional coordinates
        ''')

    x_onetep_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        onetep_cell_volume
        ''')

    x_onetep_optimised_atom_labels = Quantity(
        type=str,
        shape=['number_of_atoms'],
        description='''
        Temporary storing atomic positions
        ''')

    x_onetep_optimised_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Storing atomic optimised positions in fractional coordinates
        ''')

    x_onetep_velocities_cell_vector = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        cell vector velocities
        ''')

    x_onetep_number_of_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number_of_atoms
        ''')

    x_onetep_number_of_ngwf = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number_of_ngwf
        ''')

    x_onetep_number_of_projectors = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number_of_proj
        ''')

    x_onetep_store_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom labels
        ''')

    x_onetep_store_atom_positions = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_onetep_store_atom_ionic_velocities = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_onetep_atom_ionic_velocities = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Temporary storing atomic positions
        ''')

    x_onetep_store_optimised_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_onetep_store_optimised_atom_positions = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_onetep_section_atom_ionic_velocities = SubSection(
        sub_section=SectionProxy('x_onetep_section_atom_ionic_velocities'),
        repeats=True)

    x_onetep_section_atom_positions_optim = SubSection(
        sub_section=SectionProxy('x_onetep_section_atom_positions_optim'),
        repeats=True)

    x_onetep_section_atom_positions = SubSection(
        sub_section=SectionProxy('x_onetep_section_atom_positions'),
        repeats=True)

    x_onetep_section_cell_optim = SubSection(
        sub_section=SectionProxy('x_onetep_section_cell_optim'),
        repeats=True)

    x_onetep_section_cell = SubSection(
        sub_section=SectionProxy('x_onetep_section_cell'),
        repeats=True)


class BasisSetCellDependent(simulation.method.BasisSetCellDependent):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_basis_set_planewave_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing plane wave cutoff as string
        ''')

    x_onetep_size_std_grid = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        size of standard grid (eV)
        ''')

    x_onetep_size_fine_grid = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        sise of fine grid (1/A)
        ''')


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_enthalpy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        onetep_Enthalpy
        ''')

    x_onetep_frequency = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        onetep_frequency (cm-1)
        ''')

    x_onetep_improved_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_forces
        ''')

    x_onetep_frame_time_0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        onetep_store_t_md_frame
        ''')

    x_onetep_geom_iteration_index = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Index for number of iterations in geometry optimisation
        ''')

    x_onetep_pseudo_local = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Pseudopotential
        ''')

    x_onetep_pseudo_non_local = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Pseudopotential
        ''')

    x_onetep_rms_gradient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        RMS Gradient
        ''')

    x_onetep_n_ngwf_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        NGWF iterations
        ''')

    x_onetep_integrated_density = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        integrated density
        ''')

    x_onetep_dispersion_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Dispersion correction to the final SCF energy
        ''')

    x_onetep_ewald_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Ewald correction
        ''')

    x_onetep_store_atom_forces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_onetep_store_atom_ionforces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_onetep_store_atom_localforces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_onetep_store_atom_nonlocalforces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_onetep_store_atom_nonselfforces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_onetep_store_atom_corrforces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_onetep_initial_scf_iteration_wall_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Initial SCF iteration wall time
        ''')

    x_onetep_total_dispersion_corrected_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total electronic energy that includes dispersion energy computed with
        Disp_method_name not corrected for finite basis-set
        ''')

    x_onetep_total_energy_corrected_for_finite_basis = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        onetep_total_energy_corrected_for_finite_basis
        ''')

    x_onetep_total_energy_corrected_for_finite_basis_store = Quantity(
        type=str,
        shape=[],
        description='''
        onetep_total_energy_corrected_for_finite_basis_store
        ''')

    x_onetep_number_of_scf_iterations_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of scf iterations in single point calculation
        ''')

    x_onetep_energy_total_scf_iteration_list = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        Total electronic energy calculated with XC_method_scf during the scf iterations is
        stored in a list
        ''')

    x_onetep_ts_coordinate_path = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TS coordinate path
        ''')

    x_onetep_energy_reac = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy of reactant
        ''')

    x_onetep_energy_prod = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy of product
        ''')

    x_onetep_geom_optim_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy of product
        ''')

    x_onetep_energy_lst_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy of LST max
        ''')

    x_onetep_location_lst_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        location of LST max
        ''')

    x_onetep_barrier_reac = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        barrier from reac
        ''')

    x_onetep_barrier_prod = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        barrier from prod
        ''')

    x_onetep_reaction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        reaction energy
        ''')

    x_onetep_section_scf_k_points = SubSection(
        sub_section=SectionProxy('x_onetep_section_scf_k_points'),
        repeats=True)

    x_onetep_section_spin_number = SubSection(
        sub_section=SectionProxy('x_onetep_section_spin_number'),
        repeats=True)

    x_onetep_section_stress_tensor = SubSection(
        sub_section=SectionProxy('x_onetep_section_stress_tensor'),
        repeats=True)

    x_onetep_section_tddft = SubSection(sub_section=x_onetep_section_tddft.m_def, repeats=True)


class Energy(simulation.calculation.Energy):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_pseudopotential_local = SubSection(sub_section=simulation.calculation.EnergyEntry.m_def)

    x_onetep_pseudopotential_non_local = SubSection(sub_section=simulation.calculation.EnergyEntry.m_def)


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_functional_and_weight = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional+weight in onetep convention
        ''')

    x_onetep_section_functional_definition = SubSection(
        sub_section=SectionProxy('x_onetep_section_functional_definition'),
        repeats=True)

    x_onetep_section_functionals = SubSection(
        sub_section=SectionProxy('x_onetep_section_functionals'),
        repeats=True)

    x_onetep_section_relativity_treatment = SubSection(
        sub_section=SectionProxy('x_onetep_section_relativity_treatment'),
        repeats=True)

    x_onetep_input_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        '''
    )


class GeometryOptimization(workflow.GeometryOptimization):

    m_def = Section(validate=False, extends_base_section=True)

    geometry_optimization_frequency_tol = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Specifies the estimated average phonon frequency (as an energy) used to initialize
        the inverse Hessian matrix for geometry optimization. (eV))
        ''')

    x_onetep_max_number_of_steps = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number_of iterations geom_optim
        ''')

    geometry_optimization_geometry_conv_win = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        geom optim number of consecutive iterations during which convergence must be met
        ''')


class MolecularDynamics(workflow.MolecularDynamics):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        thermostat_target_temperature(K)
        ''')

    x_onetep_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        barostat_type
        ''')

    x_onetep_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        thermostat_type
        ''')

    x_onetep_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        thermostat_type
        ''')

    x_onetep_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        barostat_tau
        ''')

    x_onetep_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_time_step (ps)
        ''')

    x_onetep_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_time_step_number
        ''')

    x_onetep_frame_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        MD_pressure
        ''')

    x_onetep_frame_energy_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_scf_energy tolerance (eV)
        ''')

    x_onetep_frame_eigen_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_scf_eigen tolerance (eV)
        ''')


class BandStructure(simulation.calculation.BandStructure):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_k_path = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_onetep_store_k_path = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_onetep_section_eigenvalues_1 = SubSection(
        sub_section=SectionProxy('x_onetep_section_eigenvalues_1'),
        repeats=True)

    x_onetep_section_eigenvalues = SubSection(
        sub_section=SectionProxy('x_onetep_section_eigenvalues'),
        repeats=True)

    x_onetep_section_k_band = SubSection(
        sub_section=SectionProxy('x_onetep_section_k_band'),
        repeats=True)

    x_onetep_section_k_points_1 = SubSection(
        sub_section=SectionProxy('x_onetep_section_k_points_1'),
        repeats=True)

    x_onetep_section_k_points = SubSection(
        sub_section=SectionProxy('x_onetep_section_k_points'),
        repeats=True)


class ScfIteration(simulation.calculation.ScfIteration):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_scf_rms_gradient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        RMS Gradient
        ''')


class AtomParameters(simulation.method.AtomParameters):

    m_def = Section(validate=False, extends_base_section=True)

    x_onetep_store_atom_mass = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom mass
        ''')

    x_onetep_n_ngwf_atom_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number ngwf per atom
        ''')

    x_onetep_ngwf_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        radius ngwf
        ''')

    x_onetep_n_ngwf_atom = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number ngwf per atom
        ''')

    x_onetep_ngwf_radius_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        radius ngwf
        ''')

    x_onetep_store_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom name
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class x_lib_atoms_section_gap(MSection):
    '''
    Description of Gaussian Approximation Potentials (GAPs).
    '''

    m_def = Section(validate=False)

    x_lib_atoms_training_config_refs = Quantity(
        type=simulation.calculation.Calculation,
        shape=['n_sparseX'],
        description='''
        References to frames in training configuration.
        ''')

    x_lib_atoms_GAP_params_label = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_GAP_params_svn_version = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_GAP_data_do_core = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_GAP_data_e0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_command_line_command_line = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpSparse_n_coordinate = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_n_permutations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_sparsified = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_signal_variance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_label = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_n_sparseX = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_covariance_type = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_signal_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_sparseX_filename = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_dimensions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_theta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_descriptor = Quantity(
        type=str,
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_perm_permutation = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_perm_i = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_alpha = Quantity(
        type=np.dtype(np.float64),
        shape=['n_sparseX', 2],
        description='''
        GAP classifier.
        ''')

    x_lib_atoms_gpCoordinates_sparseX = Quantity(
        type=np.dtype(np.float64),
        shape=['n_sparseX', 'dimensions'],
        description='''
        GAP classifier.
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_lib_atoms_section_gap = SubSection(
        sub_section=SectionProxy('x_lib_atoms_section_gap'),
        repeats=False)


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_lib_atoms_virial_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='pascal',
        description='''
        Virial tensor for this frame.
        ''')

    x_lib_atoms_config_type = Quantity(
        type=str,
        shape=[],
        description='''
        Configuration type, e.g. = dislocation_quadrupole.
        ''')
